from marshmallow import Schema, fields, post_load
from ..resource import Resource
from collections import namedtuple


class Subscription(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#list-customer-subscriptions
    https://dev.chartmogul.com/v1.0/reference#list-a-customers-subscriptions
    """
    _path = "/customers{/uuid}/subscriptions"
    _root_key = 'entries'
    _many = namedtuple('Subscriptions', [_root_key, "has_more", "per_page", "page"])

    class _Schema(Schema):
        id = fields.Int(allow_none=True)
        plan = fields.String(allow_none=True)
        quantity = fields.Int(allow_none=True)
        mrr = fields.Number(allow_none=True)
        arr = fields.Number(allow_none=True)
        status = fields.String(allow_none=True)
        billing_cycle = fields.String(load_from='billing-cycle', allow_none=True)
        billing_cycle_count = fields.Number(load_from='billing-cycle-count', allow_none=True)
        start_date = fields.DateTime(load_from='start-date', allow_none=True)
        end_date = fields.DateTime(load_from='end-date', allow_none=True)
        currency = fields.String(allow_none=True)
        currency_sign = fields.String(load_from='currency-sign', allow_none=True)

        # /import namespace
        uuid = fields.String(allow_none=True)
        external_id = fields.String(allow_none=True)
        subscription_set_external_id = fields.String(allow_none=True)
        plan_uuid = fields.String(allow_none=True)
        customer_uuid = fields.String(allow_none=True)
        data_source_uuid = fields.String(allow_none=True)
        cancellation_dates = fields.List(fields.DateTime(), allow_none=True)

        @post_load
        def make(self, data):
            return Subscription(**data)

    _schema = _Schema(strict=True)

    # /import has different paging
    @classmethod
    def _loadJSON(cls, jsonObj):
        if "subscriptions" in jsonObj:
            _many = namedtuple('Subscriptions', ["subscriptions", "current_page", "total_pages", "customer_uuid"])
            return _many(cls._schema.load(jsonObj["subscriptions"], many=True).data,
                         jsonObj["current_page"],
                         jsonObj["total_pages"],
                         jsonObj["customer_uuid"])
        else:
            return super(Subscription, cls)._loadJSON(jsonObj)

# /import namespace
Subscription.list_imported = Subscription._method('list_imported', 'get', "/import/customers{/uuid}/subscriptions")
Subscription.cancel = Subscription._method('cancel', 'patch', "/import/subscriptions{/uuid}")
Subscription.modify = Subscription._method('modify', 'patch', "/import/subscriptions{/uuid}")



from ..api.subscription import Subscription as SubsNew
from warnings import warn


class Subscription(SubsNew):
    @classmethod
    def all(cls, *args, **kwargs):
        warn("chartmogul.imp namespace is deprecated, use chartmogul.Subscription.list_imported!")
        return super(Subscription, cls).list_imported(*args, **kwargs)

    @classmethod
    def cancel(cls, *args, **kwargs):
        warn("chartmogul.imp namespace is deprecated, use chartmogul.Subscription.cancel!")
        return super(Subscription, cls).cancel(*args, **kwargs)



VERSION = "v1"
API_BASE = "https://api.chartmogul.com"


class Config:
    uri = API_BASE + "/" + VERSION

    def __init__(self, account_token, secret_key, request_timeout=None, max_retries=20, backoff_factor=2):
        self.auth = (account_token, secret_key)
        self.request_timeout = request_timeout
        self.max_retries = max_retries
        self.backoff_factor = backoff_factor



from ..api.invoice import Invoice as InvoiceNew
from warnings import warn


class Invoice(InvoiceNew):
    @classmethod
    def create(cls, *args, **kwargs):
        warn("chartmogul.imp namespace is deprecated, use chartmogul.Invoice.create!")
        return super(Invoice, cls).create(*args, **kwargs)

    @classmethod
    def all(cls, *args, **kwargs):
        warn("chartmogul.imp namespace is deprecated, use chartmogul.Invoice.all!")
        return super(Invoice, cls).all(*args, **kwargs)



from marshmallow import Schema, fields, post_load
from ..resource import Resource
from collections import namedtuple


class Activity(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#list-customer-subscriptions
    """
    _path = "/customers{/uuid}/activities"
    _root_key = 'entries'
    _many = namedtuple('Activities', [_root_key, "has_more", "per_page", "page"])

    class _Schema(Schema):
        id = fields.Int()
        activity_arr = fields.Number(load_from='activity-arr')
        activity_mrr = fields.Number(load_from='activity-mrr')
        activity_mrr_movement = fields.Number(load_from='activity-mrr-movement')
        currency = fields.String()
        currency_sign = fields.String(load_from='currency-sign')
        date = fields.DateTime()
        description = fields.String()
        type = fields.String()

        @post_load
        def make(self, data):
            return Activity(**data)

    _schema = _Schema(strict=True)



# -*- coding: utf-8 -*-
from .api.config import Config
from .errors import APIError, ConfigurationError, ArgumentMissingError

from .api.activity import Activity
from .api.attributes import Attributes
from .api.custom_attrs import CustomAttributes
from .api.customer import Customer
from .api.data_source import DataSource
from .api.invoice import Invoice
from .api.metrics import Metrics
from .api.ping import Ping
from .api.plan import Plan
from .api.subscription import Subscription
from .api.tags import Tags
from .api.transaction import Transaction

# Deprecated
import imp


"""
ChartMogul API Python Client

Provides convenient Python bindings for ChartMogul's API.

:copyright: (c) 2017 by ChartMogul Ltd.
:license: MIT, see LICENSE for more details.
"""

__title__ = 'chartmogul'
__version__ = '1.2.5'
__build__ = 0x000000
__author__ = 'ChartMogul Ltd'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 ChartMogul Ltd'



from ..api.transaction import Transaction as TransactionNew
from warnings import warn


class Transaction(TransactionNew):
    @classmethod
    def create(cls, *args, **kwargs):
        warn("chartmogul.imp namespace is deprecated, use chartmogul.Transaction.create!")
        return super(Transaction, cls).create(*args, **kwargs)



import requests
from json import dumps
from promise import Promise
from uritemplate import URITemplate
from .retry_request import requests_retry_session
from .errors import APIError, ConfigurationError, ArgumentMissingError, annotateHTTPError
from .api.config import Config
from datetime import datetime, date
from builtins import str

"""
HTTP verb mapping. Based on nodejs library.
"""
MAPPINGS = {
    'add': 'post',
    'all': 'get',
    'cancel': 'patch',
    'create': 'post',
    'destroy': 'delete',
    'merge': 'patch',
    'modify': 'patch',
    'patch': 'patch',
    'remove': 'delete',
    'retrieve': 'get',
    'update': 'put'
}

LIST_PARAMS = ['current_page', 'total_pages',
               'has_more', 'per_page', 'page',
               'summary', 'customer_uuid']
ESCAPED_QUERY_KEYS = {
    'start_date': 'start-date',
    'end_date': 'end-date'
}


class DataObject(object):

    def __init__(self, **kwargs):
        """
        Any arguments are translated into class attributes.
        """
        self._set_attrs(**kwargs)

    def _set_attrs(self, **kwargs):
        """
        Dynamically creates attributes of the class, so it's not necessary
        to repeat their list besides the serialization schema.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        """
        Pretty-prints any object as
        <ClassName{property='string',number=4,date=datetime.datetime(2003, 8, 4, 21, 41, 43)}>
        """
        return "<" + self.__class__.__name__ + "{" + \
            ", ".join([attr + "=" + repr(getattr(self, attr)) for attr in sorted(dir(self))
                       if not attr.startswith('_') and not callable(getattr(self, attr))]) + \
               "}>"


def json_serial(obj):
    """
    JSON serializer for objects not serializable by default json code
    Handles date & datetime.
    """

    if isinstance(obj, date) or isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError("Type not serializable")


class Resource(DataObject):
    _path = ''

    @classmethod
    def _load(cls, response):
        """
        Python doesn't allow arbitrary fields on arrays and having special classes
        for array & cursor attributes is not necessary => namedtuple.
        (immutable, therefore all set on creation by generator)
        """
        response.raise_for_status()
        if response.status_code == 204 or response.status_code == 202:
            return None
        try:
            jsonObj = response.json()
        except ValueError:  # Couldn't parse JSON, probably just text message.
            return response.content
        return cls._loadJSON(jsonObj)

    @classmethod
    def _loadJSON(cls, jsonObj):
        # has load_many capability & is many entries result?
        if '_root_key' in dir(cls) is not None and cls._root_key in jsonObj:
            return cls._many(cls._schema.load(jsonObj[cls._root_key], many=True).data,
                             **{key: jsonObj[key] for key in LIST_PARAMS if key in jsonObj})
        else:
            return cls._schema.load(jsonObj).data

    @classmethod
    def _preProcessParams(cls, params):
        for key, replacement in ESCAPED_QUERY_KEYS.items():
            if key in params:
                params[replacement] = params[key]
                del params[key]
        return params

    @classmethod
    def _request(cls, config, method, http_verb, path, data=None, **kwargs):
        if http_verb == 'get':
            params = cls._preProcessParams(kwargs)
            data = None
        else:
            params = None
            if data is not None:
                data = dumps(data, default=json_serial)

        return Promise(lambda resolve, _:
            resolve(getattr(requests_retry_session(config.max_retries, config.backoff_factor), http_verb)(
                config.uri + path,
                data=data,
                headers={'content-type': 'application/json'},
                params=params,
                auth=config.auth,
                timeout=config.request_timeout)
            )).then(cls._load).catch(annotateHTTPError)

    @classmethod
    def _expandPath(cls, path, kwargs):
        t = URITemplate(path)
        return t.expand(kwargs)

    @classmethod
    def _method(cls, method, http_verb, path=None):
        @classmethod
        def fc(cls, config, **kwargs):
            if config is None or not isinstance(config, Config):
                raise ConfigurationError("First argument should be"
                                         " instance of chartmogul.Config class!")

            pathTemp = path  # due to Python closure
            if pathTemp is None:
                pathTemp = cls._path

            # This enforces user to pass argument, otherwise we could call
            # wrong URL.
            if method in ['destroy', 'cancel', 'retrieve', 'update'] and 'uuid' not in kwargs:
                raise ArgumentMissingError("Please pass 'uuid' parameter")
            if method in ['create', 'modify'] and 'data' not in kwargs:
                raise ArgumentMissingError("Please pass 'data' parameter")

            pathTemp = Resource._expandPath(pathTemp, kwargs)
            # UUID is always path parameter only.
            if 'uuid' in kwargs:
                del kwargs['uuid']

            return cls._request(config, method, http_verb, pathTemp, **kwargs)
        return fc

def _add_method(cls, method, http_verb, path=None):
    """
    Dynamically define all possible actions.
    """
    fc = Resource._method(method, http_verb, path)
    # Not supported by 2.7, but probably not needed.
    # fc.__doc__ = "Sends %s request to ChartMogul." % http_verb
    # fc.__name__ = method
    setattr(cls, method, fc)


for method, http_verb in MAPPINGS.items():
    _add_method(Resource, method, http_verb)
